/*
 * pm_transformed_task.h
 *
 *  Created on: 13.10.2015
 *      Author: Raphael Imahorn <raphael.imahorn@stud.unibas.ch>
 */

#ifndef PM_TRANSFORMED_TASK_H_
#define PM_TRANSFORMED_TASK_H_

#include "delegating_task.h"

#include "generated_operator.h"
#include "task_tools.h"
#include "task_proxy.h"

#include <unordered_map>
#include <string>
#include <list>
#include <vector>
#include <utility>


class PmTransformedTask: public DelegatingTask {
    typedef std::vector<int> AtomCollection;
public:
	PmTransformedTask(const Options &opts);
	virtual ~PmTransformedTask() override = default;

	virtual int get_num_variables() const override;
    virtual const std::string get_variable_name(int var) const override;
    virtual int get_variable_domain_size(int var) const override;
    virtual const std::string get_fact_name(int var, int value) const override;

    virtual int get_operator_cost(int index, bool is_axiom) const override;
    virtual const std::string &get_operator_name(int index, bool is_axiom) const override;
    virtual int get_num_operators() const override;
    virtual int get_num_operator_preconditions(int index, bool is_axiom) const override;
    virtual std::pair<int, int> get_operator_precondition(
        int op_index, int fact_index, bool is_axiom) const override;
    virtual int get_num_operator_effects(int op_index, bool is_axiom) const override;
    virtual int get_num_operator_effect_conditions(
        int op_index, int eff_index, bool is_axiom) const override;
    virtual std::pair<int, int> get_operator_effect_condition(
        int op_index, int eff_index, int cond_index, bool is_axiom) const override;
    virtual std::pair<int, int> get_operator_effect(
        int op_index, int eff_index, bool is_axiom) const override;
    virtual const GlobalOperator *get_global_operator(int index, bool is_axiom) const override;

    virtual int get_num_goals() const override;
    virtual std::pair<int, int> get_goal_fact(int index) const override;

    virtual std::vector<int> get_initial_state_values() const override;
    virtual std::vector<int> get_state_values(const GlobalState &global_state) const override;
private:
	void init();
	void init_atoms();
	void init_meta_atoms();
	void init_findmap();
	void init_operators();
	void init_initial_state();
	void init_goals();
	void add_new_meta_atom(const AtomCollection &atoms);
	void add_new_operator(OperatorProxy &orig_op, const AtomCollection &meta_atom);
	void append_k_sized_meta_atom_subset(int k);
	void exit_cause_axiom() const;
	void exit_cause_effect_conditions(int op_index) const;
	std::vector<int>* index_subset_of_size_k(const AtomCollection &set, int subset_size,
			std::vector<int> *out);
	void index_subset_of_size_k(const AtomCollection &set, int subset_size,
			AtomCollection &l, AtomCollection::const_iterator it, std::vector<int> *out);
	std::vector<AtomCollection> subset_of_size_k(const AtomCollection &set, int subset_size);
	void subset_of_size_k(const AtomCollection &set, int subset_size,
			AtomCollection &l, AtomCollection::const_iterator it, std::vector<AtomCollection> *out);
	void print_atom_collection(const AtomCollection &l, char end_char = '\n', bool new_line = false);
	bool check_disjoint(const EffectsProxy &effects, const AtomCollection &meta_atom);
	bool contains_at_least_one_list_element(const AtomCollection &which, const AtomCollection &what);
	AtomCollection remove_doubles(const AtomCollection &in);
	AtomCollection get_merged(const AtomCollection &lhs, const AtomCollection &rhs);
	int get_atom(int var_index, int fact_index) const;
	int bin(int n, int k);
	std::vector<int> make_effects(const OperatorProxy &orig_op, const AtomCollection &meta_atom);
	std::vector<int> make_preconditions(const OperatorProxy &orig_op, const AtomCollection &meta_atom);
	std::string get_atom_fact_name(int atom, int fact = 1) const;
	std::string get_meta_atom_fact_name(int meta_atom) const;


	std::vector<GeneratedOperator*> operators;
	std::vector<int> parent_operator;

	std::vector<std::pair<int,int>> atoms;
	std::vector<int> atom_helper;

	// TODO besser wenn durch Pointer ersetzt?
	std::vector<AtomCollection> meta_atoms;

	struct hash_func{
		size_t operator()(const PmTransformedTask::AtomCollection * ac) const {
			return hash_sequence(*ac, ac->size());
		}
	};
	struct equal_func{
	    bool operator()(const PmTransformedTask::AtomCollection *ac1,
	    		const PmTransformedTask::AtomCollection *ac2) const {
	        return *ac1 == *ac2;
	    }
	};
	typedef std::unordered_map<AtomCollection*, int, hash_func, equal_func> FindMap;
	FindMap findmap;

	std::vector<int> goals;

	std::vector<int> initial_state_values;

	const char SEPARATOR = '.';
	const std::string NEGATION_PREFIX = "NOT ";
	const std::string META_ATOM_JOINER = " AND ";
	unsigned int m;
	unsigned parent_num_facts = 0;
	int parent_num_variables;
	int parent_num_operators;

	TaskProxy parentProxy;
};


#endif /* PM_TRANSFORMED_TASK_H_ */
